# vim: set ts=4 sw=4 et: coding=UTF-8

from .rpmsection import Section


class RpmDescription(Section):

    """
    Only keep one empty line for many consecutive ones.
    Remove Authors from description.
    """

    def __init__(self, options):
        Section.__init__(self, options)
        self.removing_authors = False
        # Tracks the use of a macro. When this happens and we're still in a
        # description, we actually don't know where we are so we just put all
        # the following lines blindly, without trying to fix anything.
        self.unknown_line = False

    def add(self, line):
        if self.previous_line and len(line) > 0 and line[0] == '%':
            self.unknown_line = True

        if self.removing_authors and not self.unknown_line:
            return

        if len(line) == 0:
            if not self.previous_line or len(self.previous_line) == 0:
                return

        if self.reg.re_authors.match(line) and not self.minimal:
            self.removing_authors = True
            return

        Section.add(self, line)



"""RPM dependency lines parser and helpers.

Contains class DependencyParser which parses string and generates
token tree. For common manipulation is method flat_out() useful, it
just splits dependencies into list.

For future development is useful find_end_of_macro().

"""
import re
import logging

from .rpmexception import NoMatchException

DEBUG = None

if DEBUG:
    logging.basicConfig(level=logging.DEBUG)

re_parens = re.compile(
    r'(' +
    r'\('  + r'|' + r'\)'  + r'|' +
    r'\\(' + r'|' + r'\\)' + r'|' +
    r'[^\()]+' +
    r')'
)

re_braces = re.compile(
    r'(' +
    r'\{'  + r'|' + r'\}'  + r'|' +
    r'\\{' + r'|' + r'\\}' + r'|' +
    r'[^\{}]+' +
    r')'
)

re_name = re.compile(r'[-A-Za-z0-9_~():.+/*]+')
re_version = re.compile(r'[-A-Za-z0-9_~():.+]+')
re_spaces = re.compile(r'\s+')
re_macro_unbraced = re.compile('%[A-Za-z0-9_]{3,}')
re_version_operator = re.compile('(>=|<=|=>|=<|>|<|=)')

def find_end_of_macro(string, regex, opening, closing):
    if DEBUG:
        logger = logging.getLogger('DepParser')
    else:
        logger = None
    macro = string[0:2]
    # eat '%{'
    string = string[2:]

    opened = 1
    while opened and string:
        if logger:
            logger.debug('opened: %d string: %s', opened, string)
        try:
            bite, string = consume_chars(regex, string, logger)
        except NoMatchException:
            raise Exception('unexpected parser error when looking for end of '
                            'macro')

        if bite == opening:
            opened += 1
        elif bite == closing:
            opened -= 1
        macro += bite

    if opened:
        raise Exception('Unexpectedly met end of string when looking for end '
                        'of macro')
    return macro


def consume_chars(regex, string, logger=None):
    if logger:
        logger.debug('consume_chars: regex: "%s"', regex.pattern)
        logger.debug('consume_chars: string:"%s"', string)
    match = regex.match(string)
    if match:
        end = match.end()
        if logger:
            logger.debug('consume_chars: split "%s", "%s"', string[0:end], string[end:])
        return string[0:end], string[end:]
    else:
        raise NoMatchException('Expected match failed')


class DependencyParser(object):

    logger = None

    def __init__(self, string):
        self.string = string.rstrip()
        self.token = []
        self.parsed = []
        self.state = ['name']
        if DEBUG:
            self.logger = logging.getLogger('DepParser')
            self.logger.setLevel(logging.DEBUG)
        self.state_change_loop()

    def dump_token(self):
        if self.logger:
            self.logger.debug('dump_token')
        self.status()
        if not self.token:
            return
        if self.token[0].isspace():
            self.token = self.token[1:]
            if not self.token:
                return
        self.parsed.append(self.token)
        self.token = []

    def state_change_loop(self):
        while self.string:
            if self.state[-1] == 'name':
                self.read_name()
            elif self.state[-1] == 'version_operator':
                self.read_version_operator()
            elif self.state[-1] == 'version':
                self.read_version()
            elif self.state[-1] == 'macro_name':
                self.read_macro_name()
            elif self.state[-1] == 'macro_shell':
                self.read_macro_shell()
            elif self.state[-1] == 'macro_unbraced':
                self.read_macro_unbraced()
            elif self.state[-1] == 'spaces':
                self.read_spaces()
        self.dump_token()

    def status(self):
        if self.logger:
            self.logger.debug('token: %s', self.token)
            self.logger.debug('string: "%s"', self.string)
            self.logger.debug('parsed: %s', self.parsed)
            self.logger.debug('state: %s', self.state)
            self.logger.debug('--------------------------------')

    def read_spaces(self, state_change=True):
        try:
            spaces, self.string = consume_chars(re_spaces, self.string, self.logger)
            self.token.append(spaces)
            if state_change:
                self.state.pop()  # remove 'spaces' state
                # if we were reading version, space definitely means
                # end of that
                if self.state[-1] == 'version':
                    self.state.pop()
                    self.dump_token()
            self.status()
        except NoMatchException:
            pass

    def read_unknown(self):
        '''
        Try to identify, what is to be read now.
        '''
        if self.string[0:2] in ['>=', '<=', '=>', '=<'] or \
                self.string[0:1] in ['<', '>', '=']:
            self.state.append('version')
            self.state.append('version_operator')
        elif self.string[0] == ' ':
            self.state.append('spaces')
        elif self.string[0:2] == '%{':
            self.state.append('macro_name')
        elif self.string[0:2] == '%(':
            self.state.append('macro_shell')
        elif self.string[0:2] == '%%':
            self.read_double_percent()
        elif self.string[0] == '%':
            self.state.append('macro_unbraced')
        elif self.string[0] == ',':
            self.string = self.string[1:]
            self.dump_token()
        if self.logger:
            self.logger.debug('read_unknown: states: %s string: "%s"',
                              self.state, self.string)

    def read_name(self):
        try:
            name, self.string = consume_chars(re_name, self.string, self.logger)
            if self.token and self.token[-1].isspace():
                self.dump_token()
            self.token.append(name)
            self.status()
        except NoMatchException:
            self.read_unknown()

    def read_double_percent(self):
        self.token.append('%%')
        self.string = self.string[2:]

    def read_macro_unbraced(self):
        try:
            # 3 or more alphanumeric characters
            macro, self.string = consume_chars(
                re_macro_unbraced, self.string, self.logger)
            self.token.append(macro)
            self.state.pop()  # remove 'macro_unbraced' state
            self.status()
        except NoMatchException:
            self.read_unknown()

    def read_version_operator(self):
        try:
            operator, self.string = consume_chars(
                re_version_operator, self.string, self.logger)
            self.token.append(operator)
            # Note: this part is a bit tricky, I need to read possible
            # spaces or tabs now so I won't get to [ ..., 'version',
            # 'spaces' ] state before the end
            self.read_spaces(state_change=False)
            self.state.pop()  # get rid of 'version_operator'
            self.status()
        except NoMatchException:
            self.read_unknown()

    def read_version(self):
        try:
            version, self.string = consume_chars(
                re_version, self.string, self.logger)
            self.token.append(version)
            self.status()
        except NoMatchException:
            self.read_unknown()

    def read_macro_name(self):
        macro = find_end_of_macro(self.string, re_braces, '{', '}')
        # remove macro from string
        self.string = self.string[len(macro):]
        self.token.append(macro)
        # now we expect previous state
        self.state.pop()
        self.status()

    def read_macro_shell(self):
        macro = find_end_of_macro(self.string, re_parens, '(', ')')
        self.string = self.string[len(macro):]
        self.token.append(macro)
        # now we expect previous state
        self.state.pop()
        self.status()

    def flat_out(self):
        result = []
        for token in self.parsed:
            if isinstance(token, list):
                if token and token[-1].isspace():
                    token = token[:-1]
                result.append(''.join(token))
            else:
                if not token.isspace():
                    result.append(token)
        return result



# vim: set ts=4 sw=4 et: coding=UTF-8

# Copyright (c) 2015, SUSE LINUX Products GmbH, Nuernberg, Germany
# All rights reserved.
# See COPYING for details.

import os
import sys
import argparse

from .rpmexception import RpmWrongArgs, RpmException
from .rpmcleaner import RpmSpecCleaner


__version__ = '0.9.3'


def process_args(argv):
    """
    Process the parsed arguments and return the result
    :param argv: passed arguments
    """

    parser = argparse.ArgumentParser(prog='spec-cleaner',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description='Cleans the given spec file according to style guide and returns the result.')

    # Make the -d, -i, and -o exclusive as we can do only one of those
    output_group = parser.add_mutually_exclusive_group()

    parser.add_argument('spec', metavar='SPEC', type=str,
                        help='spec file to beautify')
    parser.add_argument('-c', '--cmake', action='store_true',
                        help='convert dependencies to their cmake() counterparts, requires bit more of cleanup in spec afterwards.')
    output_group.add_argument('-d', '--diff', action='store_true',
                              help='run the diff program to show differences between new and original specfile.')
    parser.add_argument('--diff-prog', default='vimdiff',
                        help='specify the diff binary to call with diff option.')
    parser.add_argument('-f', '--force', action='store_true',
                        help='overwrite the output file if already exist.')
    output_group.add_argument('-i', '--inline', action='store_true',
                              help='inline the changes directly to the parsed file.')
    parser.add_argument('-m', '--minimal', action='store_true',
                        help='run in minimal mode that does not do anything intrusive (ie. just sets the Copyright)')
    parser.add_argument('--no-copyright', action='store_true',
                        help='do not include official SUSE copyright hear and just keep what is present')
    output_group.add_argument('-o', '--output', default='',
                              help='specify the output file for the cleaned spec content.')
    parser.add_argument('-p', '--pkgconfig', action='store_true',
                        help='convert dependencies to their pkgconfig() counterparts, requires bit more of cleanup in spec afterwards.')
    parser.add_argument('--perl', action='store_true',
                        help='convert dependencies to their perl() counterparts, highly expansive because most main packages provide some perl symbols, use carefully')
    parser.add_argument('-t', '--tex', action='store_true',
                        help='convert dependencies to their tex() counterparts, requires verification of the output afterwards.')
    parser.add_argument('-v', '--version', action='version', version=__version__,
                        help='show package version and exit')
    parser.add_argument('-k', '--keep-space', action='store_true',
                        help='keep empty lines in preamble intact.')

    # print help if there is no argument
    if len(argv) < 1:
        parser.print_help()
        sys.exit(0)

    options = parser.parse_args(args=argv)

    # the spec must exist for us to do anything
    if not os.path.exists(options.spec):
        raise RpmWrongArgs('{0} does not exist.'.format(options.spec))

    # the path for output must exist and the file must not be there unless
    # force is specified
    if options.output:
        options.output = os.path.expanduser(options.output)
        if not options.force and os.path.exists(options.output):
            raise RpmWrongArgs('{0} already exists.'.format(options.output))

    # convert options to dict
    options_dict = {
        'specfile': options.spec,
        'output': options.output,
        'pkgconfig': options.pkgconfig,
        'inline': options.inline,
        'diff': options.diff,
        'diff_prog': options.diff_prog,
        'minimal': options.minimal,
        'no_copyright': options.no_copyright,
        'perl': options.perl,
        'tex': options.tex,
        'cmake': options.cmake,
        'keep_space': options.keep_space,
    }

    return options_dict


def main():
    """
    Main function that calls argument parsing ensures their sanity
    and then creates RpmSpecCleaner object that works with passed spec file.
    """

    try:
        options = process_args(sys.argv[1:])
    except RpmWrongArgs as exception:
        sys.stderr.write('ERROR: {0}\n'.format(exception))
        return 1

    try:
        cleaner = RpmSpecCleaner(options)
        cleaner.run()
    except RpmException as exception:
        sys.stderr.write('ERROR: {0}\n'.format(exception))
        return 1

    return 0



# vim: set ts=4 sw=4 et: coding=UTF-8


class Section(object):

    """
    Basic object for parsing each section of spec file.
    It stores the lines in a list and remembers content of
    previous line at hand.
    The unbrace_keywords content is passed from creating
    object to reduce calculation price.
    Various functions do replacement of common typos to
    unificate all the content.
    """

    def __init__(self, options):
        self.lines = []
        self.previous_line = None
        self.spec = options['specfile']
        self.minimal = options['minimal']
        self.reg = options['reg']
        # Are we inside of conditional or not
        self.condition = False
        self._condition_counter = 0

    def _complete_cleanup(self, line):
        """
        Function that just calls all the cleanups in proper order so it
        can be called beforehand if we override add phase and want to
        do our replaces after cleanup.
        """

        line = line.rstrip()

        if not line.startswith('#'):
            if not self.minimal:
                line = self.embrace_macros(line)
            line = self.replace_buildroot(line)
            line = self.replace_optflags(line)
            line = self.replace_known_dirs(line)
            line = self.replace_utils(line)
            line = self.replace_buildservice(line)
            line = self.replace_preamble_macros(line)

        return line

    def _check_conditions(self, line):
        """
        Check if we are in condition that is contained or not
        """
        if self.reg.re_if.match(line) or self.reg.re_codeblock.match(line):
            self._condition_counter += 1
        if self.reg.re_endif.match(line) or self.reg.re_endcodeblock.match(line):
            self._condition_counter -= 1

        if self._condition_counter > 0:
            self.condition = True
        else:
            self.condition = False

    def add(self, line):
        """
        Run the cleanup and add the line to the list of lines
        """

        line = self._complete_cleanup(line)

        # condtions detect
        self._check_conditions(line)

        # append to the file
        self.lines.append(line)
        self.previous_line = line

    def output(self, fout, newline=True, new_class=None):
        # Always append one empty line at the end if it is not present
        # and changelog is trailing part of our spec so do not put nothing
        # bellow
        # Also if we are jumping away just after writing one macroed line
        # we don't want to create new line
        if newline and len(self.lines) >= 1:
            if self.lines[-1] != '' and \
               self.lines[-1] != '%changelog' and not \
               self.lines[-1].startswith('%if') and not \
               self.lines[-1].startswith('%pre') and not \
               self.lines[-1].startswith('%post'):
                self.lines.append('')
            if new_class != 'RpmScriptlets' and \
               (self.lines[-1].startswith('%pre') or
                    self.lines[-1].startswith('%post')):
                self.lines.append('')
            # remove the newlines around ifs if they are not wanted
            if len(self.lines) >= 2:
                if self.lines[-1] == '' and \
                   (self.lines[-2].startswith('%if') or
                        self.lines[-2].startswith('%else')):
                    self.lines.pop()

        for line in self.lines:
            fout.write(line + '\n')

    def strip_useless_spaces(self, line):
        """
        Function to remove useless multiple spaces in some areas.
        It can't be called everywhere so we have to call it in
        children classes where fit.
        """
        return ' '.join(line.split())

    def embrace_macros(self, line):
        """
        Add {} around known macros that have no arguments and are not
        on whitelist.
        Whitelist is passed from caller object
        """

        # I don't think that this can be done within one regexp replacement
        # if you have idea, send me a patch :)

        # work only with non-commented part
        sp = line.split('#')
        # so, for now, put braces around everything, what looks like macro,
        previous = sp[0]

        while True:
            sp[0] = self.reg.re_macro.sub(r'\1%{\3}\5', sp[0])
            if sp[0] == previous:
                break
            else:
                previous = sp[0]

        # and replace back known keywords to braceless state again
        sp[0] = self.reg.re_unbrace_keywords.sub(r'%\1', sp[0])
        # re-create the line back
        return '#'.join(sp)

    def replace_buildroot(self, line):
        """
        Replace RPM_BUILD_ROOT for buildroot
        """
        line = self.reg.re_rpmbuildroot.sub(r'%{buildroot}\2', line)
        line = self.reg.re_rpmbuildroot_quotes.sub(r'%{buildroot}', line)
        return line

    def replace_optflags(self, line):
        """
        Replace RPM_OPT_FLAGS for %{optflags}
        """
        # if the optflags is the only thing then also add quotes around it
        line = self.reg.re_optflags_quotes.sub('="%{optflags}"', line)
        line = self.reg.re_optflags.sub('%{optflags}', line)
        return line

    def replace_known_dirs(self, line):
        """
        Replace hardcoded stuff like /usr/share -> %{_datadir}
        """

        line = self.reg.re_oldprefix.sub(r'%{_prefix}\1', line)
        line = self.reg.re_prefix.sub(r'%{_prefix}\1', line)
        line = self.reg.re_bindir.sub(r'%{_bindir}\1', line)
        line = self.reg.re_sbindir.sub(r'%{_sbindir}\1', line)
        # Disabled on purpose as it causes conflict in fedora/opensuse
        # Unfortunate but there is no reliable workaround
        # line = self.reg.re_libexecdir.sub(r'%{_libexecdir}\1', line)
        line = self.reg.re_includedir.sub(r'%{_includedir}\1', line)
        line = self.reg.re_datadir.sub(r'%{_datadir}\1', line)
        line = self.reg.re_mandir.sub(r'%{_mandir}\1', line)
        line = self.reg.re_infodir.sub(r'%{_infodir}\1', line)
        line = self.reg.re_docdir.sub(r'%{_docdir}\1', line)
        line = self.reg.re_initdir.sub(r'%{_initddir}\1', line)
        line = self.reg.re_sysconfdir.sub(r'%{_sysconfdir}\1', line)
        line = self.reg.re_localstatedir.sub(r'%{_localstatedir}\1', line)
        line = self.reg.re_libdir.sub(r'%{_libdir}\2', line)
        line = self.reg.re_initddir.sub(r'%{_initddir}\1', line)

        return line

    def replace_utils(self, line):
        """
        Remove the macro calls for utilities and rather use direct commands.
        OBS ensures there is only one anyway.
        """
        r = {'id_u': 'id -u', 'ln_s': 'ln -s', 'lzma': 'xz --format-lzma', 'mkdir_p': 'mkdir -p',
             'awk': 'gawk', 'cc': 'gcc', 'cpp': 'gcc -E', 'cxx': 'g++', 'remsh': 'rsh', }
        for i in r:
            line = line.replace('%{__' + i + '}', r[i])
            if self.minimal:
                line = line.replace('%__' + i, r[i])

        for i in ['aclocal', 'ar', 'as', 'autoconf', 'autoheader', 'automake', 'bzip2', 'cat', 'chgrp', 'chmod', 'chown', 'cp', 'cpio', 'file', 'gpg', 'grep', 'gzip', 'id', 'install', 'ld', 'libtoolize', 'make', 'mkdir', 'mv', 'nm', 'objcopy', 'objdump', 'patch', 'perl', 'python', 'python2', 'python3', 'pypy3', 'ranlib', 'restorecon', 'rm', 'rsh', 'sed', 'semodule', 'ssh', 'strip', 'tar', 'unzip', 'xz']:
            line = line.replace('%{__' + i + '}', i)
            if self.minimal:
                line = line.replace('%__' + i, i)

        return line

    def replace_buildservice(self, line):
        """
        Pretty format the conditions for distribution/version detection.
        Replace %{suse_version} for 0%{?suse_version}
        """
        for i in ['centos', 'debian', 'fedora', 'mandriva', 'meego', 'rhel', 'sles', 'suse', 'ubuntu']:
            line = line.replace('%{' + i + '_version}', '0%{?' + i + '_version}').replace('00%{', '0%{')
        return line

    def replace_preamble_macros(self, line):
        """
        Replace %{S:0} for %{SOURCE0} and so on.
        """
        line = self.reg.re_ptch.sub(r'%{PATCH\1}', line)
        line = self.reg.re_src.sub(r'%{SOURCE\1}', line)
        return line



# vim: set ts=4 sw=4 et: coding=UTF-8

import os
import sys
import sysconfig

from .rpmexception import RpmException


class FileUtils(object):

    """
    Class working with file operations.
    Read/write..
    """

    # file variable
    f = None

    def open_datafile(self, name):
        """
        Function to open data files.
        Used all around so kept glob here for importing.
        """

        homedir = os.getenv('HOME', '~') + '/.local/'

        possible_paths = [
            '{0}/../data/{1}'.format(os.path.dirname(os.path.realpath(__file__)), name),
            '{0}/share/spec-cleaner/{1}'.format(homedir, name),
            '{0}/share/spec-cleaner/{1}'.format(sysconfig.get_path('data'), name),
            '{0}/share/spec-cleaner/{1}'.format(sys.prefix, name),
        ]
        for path in possible_paths:
            try:
                _file = open(path, 'r')
            except IOError:
                pass
            else:
                self.f = _file
                return
        # file not found
        raise RpmException("File '{0}' not found in datadirs".format(name))

    def open(self, name, mode):
        """
        Function to open regular files with exception handling.
        """

        try:
            _file = open(name, mode)
        except IOError as error:
            raise RpmException(str(error))

        self.f = _file

    def close(self):
        """
        Just wrapper for closing the file
        """

        if self.f:
            self.f.close()
            self.f = None

    def __del__(self):
        self.close()
        self.f = None



# vim: set ts=4 sw=4 et: coding=UTF-8

# We basically extend rpmcheck
from .rpmcheck import RpmCheck


class RpmBuild(RpmCheck):

    """
        Replace various troublemakers in build phase
    """

    def add(self, line):
        # we do not want to run suseupdateconfig, deprecated
        if self.reg.re_suseupdateconfig.search(line):
            return

        # if user uses cmake/configure directly just recommend him using the macros
        # but check on the multiline entry and do not ammend that
        if not self.minimal:
            if (not self.previous_line \
               or (not self.previous_line.startswith('#') \
               and not self.previous_line.endswith('\\'))) \
               and self.reg.re_configure.match(line):
                self.lines.append('# FIXME: you should use the %%configure macro')
            if (not self.previous_line \
               or (not self.previous_line.startswith('#') \
               and not self.previous_line.endswith('\\'))) \
               and self.reg.re_cmake.match(line):
                self.lines.append('# FIXME: you should use %%cmake macros')

        RpmCheck.add(self, line)



# vim: set ts=4 sw=4 et: coding=UTF-8

import re
import os

from .fileutils import FileUtils

LICENSES_CHANGES = 'licenses_changes.txt'
TEX_CONVERSIONS = 'tex_conversions.txt'
PKGCONFIG_CONVERSIONS = 'pkgconfig_conversions.txt'
PERL_CONVERSIONS = 'perl_conversions.txt'
CMAKE_CONVERSIONS = 'cmake_conversions.txt'
GROUPS_LIST = 'allowed_groups.txt'
BRACKETING_EXCLUDES = 'excludes-bracketing.txt'


def parse_rpm_showrc():
    macros = []

    re_rc_macrofunc = re.compile(r'^-[0-9]+[:=]\s(\w+)\(.*')
    output = os.popen('rpm --showrc')
    for line in output:
        line = line.rstrip('\n')
        found_macro = re_rc_macrofunc.sub(r'\1', line)
        if found_macro != line:
            macros += [found_macro]
    output.close()
    return macros


def load_keywords_whitelist():
    keywords = []

    files = FileUtils()
    files.open_datafile(BRACKETING_EXCLUDES)
    for line in files.f:
        keywords.append(line.rstrip('\n'))
    files.close()

    return keywords


def find_macros_with_arg(spec):
    macrofuncs = []

    re_spec_macrofunc = re.compile(r'^\s*%define\s(\w+)\(.*')
    files = FileUtils()
    files.open(spec, 'r')
    for line in files.f:
        line = line.rstrip('\n')
        found_macro = re_spec_macrofunc.sub(r'\1', line)
        if found_macro != line:
            macrofuncs += [found_macro]
    files.close()
    return macrofuncs


def read_conversion_changes(conversion_file):
    conversions = {}

    files = FileUtils()
    files.open_datafile(conversion_file)
    for line in files.f:
        # the values are split by  ': '
        pair = line.split(': ')
        conversions[pair[0]] = pair[1][:-1]
    files.close()
    return conversions


def read_tex_changes():
    return read_conversion_changes(TEX_CONVERSIONS)


def read_pkgconfig_changes():
    return read_conversion_changes(PKGCONFIG_CONVERSIONS)


def read_perl_changes():
    return read_conversion_changes(PERL_CONVERSIONS)


def read_cmake_changes():
    return read_conversion_changes(CMAKE_CONVERSIONS)


def read_licenses_changes():
    licenses = {}

    files = FileUtils()
    files.open_datafile(LICENSES_CHANGES)
    # Header starts with # first line so skip
    next(files.f)
    for line in files.f:
        # strip newline
        line = line.rstrip('\n')
        # file has format
        # correct license string<tab>known bad license string
        # tab is used as separator
        pair = line.split('\t')
        licenses[pair[1]] = pair[0]
    files.close()
    return licenses


def read_group_changes():
    groups = []

    files = FileUtils()
    files.open_datafile(GROUPS_LIST)
    # header starts with link where we find the groups
    next(files.f)
    for line in files.f:
        line = line.rstrip('\n')
        groups.append(line)
    files.close()
    return groups


def sort_uniq(seq):
    def _check_list(x):
        if isinstance(x, list):
            return True
        else:
            return False

    seen = {}
    result = []
    for item in seq:
        marker = item
        # We can have list there with comment
        # So if list found just grab latest in the sublist
        if _check_list(marker):
            marker = marker[-1]
        if marker in seen:
            # Not a list, no comment to preserve
            if not _check_list(item):
                continue
            # Here we need to preserve comment content
            # As the list is already sorted we can count on it to be
            # seen in previous run.
            # match the current and then based on wether the previous
            # value is a list we append or convert to list entirely
            prev = result[-1]
            if _check_list(prev):
                # Remove last line of the appending
                # list which is the actual dupe value
                item.pop()
                # Remove it from orginal
                prev.pop()
                # join together
                prev += item
                # append the value back
                prev.append(marker)
                result[-1] = prev
            else:
                # Easy as there was no list
                # just replace it with our value
                result[-1] = item
            continue
        seen[marker] = 1
        result.append(item)
    return result



# vim: set ts=4 sw=4 et: coding=UTF-8

import re

from .rpmsection import Section
from .rpmexception import RpmException
from .rpmhelpers import sort_uniq
from .dependency_parser import DependencyParser

class RpmPreamble(Section):

    """
        Only keep one empty line for many consecutive ones.
        Reorder lines.
        Fix bad licenses.
        Use one line per BuildRequires/Requires/etc.
        Standardize BuildRoot.

        This one is a bit tricky since we reorder things. We have a notion of
        paragraphs, categories, and groups.

        A paragraph is a list of non-empty lines. Conditional directives like
        %if/%else/%endif also mark paragraphs. It contains categories.
        A category is a list of lines on the same topic. It contains a list of
        groups.
        A group is a list of lines where the first few ones are comment lines,
        and the last one is a normal line.

        This means that comments will stay attached to one
        line, even if we reorder the lines.
    """

    category_to_key = {
        'name': 'Name',
        'version': 'Version',
        'release': 'Release',
        'license': 'License',
        'summary': 'Summary',
        # The localized summary can contain various values, so it can't be here
        'url': 'Url',
        'group': 'Group',
        'source': 'Source',
        'nosource': 'NoSource',
        'patch': 'Patch',
        'buildrequires': 'BuildRequires',
        'conflicts': 'Conflicts',
        'prereq': 'PreReq',
        'requires': 'Requires',
        'requires_eq': '%requires_eq',
        'recommends': 'Recommends',
        'suggests': 'Suggests',
        'enhances': 'Enhances',
        'supplements': 'Supplements',
        # Provides/Obsoletes cannot be part of this since we want to keep them
        # mixed, so we'll have to specify the key when needed
        'buildroot': 'BuildRoot',
        'buildarch': 'BuildArch',
    }

    categories_order = [
        'define',
        'bconds',
        'bcond_conditions',
        'name',
        'version',
        'release',
        'summary',
        'summary_localized',
        'license',
        'group',
        'url',
        'source',
        'nosource',
        'patch',
        'buildrequires',
        'requires',
        'requires_eq',
        'prereq',
        'requires_phase',  # this is Requires(pre/post/...)
        'recommends',
        'suggests',
        'enhances',
        'supplements',
        'conflicts',
        'provides_obsoletes',
        'buildroot',
        'buildarch',
        'misc',
        'build_conditions',
        'conditions',
    ]

    # categories that are sorted based on value in them
    categories_with_sorted_package_tokens = [
        'buildrequires',
        'prereq',
        'requires',
        'requires_eq',
        'recommends',
        'suggests',
        'enhances',
        'supplements',
        'conflicts',
    ]

    # categories that are sorted based on key value (eg Patch0 before Patch1)
    categories_with_sorted_keyword_tokens = [
        'source',
        'patch',
    ]

    def __init__(self, options):
        Section.__init__(self, options)
        # Old storage
        self._oldstore = []
        # Is the parsed variable multiline (ending with \)
        self.multiline = False
        # Are we inside of conditional or not
        self.condition = False
        # Is the condition with define/global variables
        self._condition_define = False
        # Is the condition based probably on bcond evaluation
        self._condition_bcond = False
        # do we want pkgconfig and others?
        self.pkgconfig = options['pkgconfig']
        self.perl = options['perl']
        self.cmake = options['cmake']
        self.tex = options['tex']
        # are we supposed to keep empty lines intact?
        self.keep_space = options['keep_space']
        # dict of license replacement options
        self.license_conversions = options['license_conversions']
        # dict of pkgconfig and other conversions
        self.pkgconfig_conversions = options['pkgconfig_conversions']
        self.perl_conversions = options['perl_conversions']
        self.cmake_conversions = options['cmake_conversions']
        self.tex_conversions = options['tex_conversions']
        # list of allowed groups
        self.allowed_groups = options['allowed_groups']
        # start the object
        self._start_paragraph()
        # initialize list of groups that need to pass over conversion fixer
        self.categories_with_package_tokens = self.categories_with_sorted_package_tokens[:]
        # these packages actually need fixing after we sent the values to
        # reorder them
        self.categories_with_package_tokens.append('provides_obsoletes')
        # license handling
        self.subpkglicense = options['subpkglicense']
        self.license = options['license']
        # pkgconfig requirement detection
        self.br_pkgconfig_required = False

        # simple categories matching
        self.category_to_re = {
            'name': self.reg.re_name,
            'version': self.reg.re_version,
            # license need fix replacment
            'summary': self.reg.re_summary,
            'url': self.reg.re_url,
            'group': self.reg.re_group,
            'nosource': self.reg.re_nosource,
            # for source, we have a special match to keep the source number
            # for patch, we have a special match to keep the patch number
            'buildrequires': self.reg.re_buildrequires,
            'conflicts': self.reg.re_conflicts,
            # for prereq we append warning comment so we don't mess it there
            'requires': self.reg.re_requires,
            'recommends': self.reg.re_recommends,
            'suggests': self.reg.re_suggests,
            'enhances': self.reg.re_enhances,
            'supplements': self.reg.re_supplements,
            # for provides/obsoletes, we have a special case because we group them
            # for build root, we have a special match because we force its value
            'buildarch': self.reg.re_buildarch,
        }

        # deprecated definitions that we no longer want to see
        self.category_to_clean = {
            'vendor': self.reg.re_vendor,
            'autoreqprov': self.reg.re_autoreqprov,
            'epoch': self.reg.re_epoch,
            'icon': self.reg.re_icon,
            'copyright': self.reg.re_copyright,
            'packager': self.reg.re_packager,
            'debugpkg': self.reg.re_debugpkg,
            'prefix': self.reg.re_preamble_prefix,
        }

    def _start_paragraph(self):
        self.paragraph = {}
        for i in self.categories_order:
            self.paragraph[i] = []
        self.current_group = []

    def start_subparagraph(self):
        # store the main content and clean up
        self._oldstore.append(self.paragraph)
        self._start_paragraph()

    def _add_group(self, group):
        """
        Actually store the lines from groups to resulting output
        """
        t = type(group)
        if t == str:
            return [group]
        elif t == list:
            x = []
            for subgroup in group:
                x += self._add_group(subgroup)
            return x
        else:
            raise RpmException('Unknown type of group in preamble: %s' % t)

    def _sort_helper_key(self, a):
        t = type(a)
        if t == str:
            key = a
        elif t == list:
            # if this is a list then all items except last are commentary
            key = a[-1]
        else:
            raise RpmException('Unknown type during sort: %s' % t)

        # Special case is the category grouping where we have to get the number in
        # after the value
        if self.reg.re_patch.match(key):
            match = self.reg.re_patch.match(key)
            key = int(match.group(2))
        elif self.reg.re_source.match(key):
            match = self.reg.re_source.match(key)
            value = match.group(1)
            if not value:
                value = '0'
            key = int(value)
        # Put brackety ()-style deps at the end of the list, after all other
        elif self.reg.re_brackety_requires.search(key):
            key = '1' + key
        else:
            key = '0' + key
        return key

    def end_subparagraph(self, endif=False):
        lines = self._end_paragraph()
        if len(self.paragraph['define']) > 0 or \
           len(self.paragraph['bconds']) > 0:
            self._condition_define = True
        self.paragraph = self._oldstore.pop(-1)
        self.paragraph['conditions'] += lines

        # If we are on endif we check the condition content
        # and if we find the defines we put it on top.
        if endif or not self.condition:
            # check if we are doing the ppc64 migration and delete it
            if not self.minimal and \
                 isinstance(self.paragraph['conditions'][0], list) and \
                 len(self.paragraph['conditions']) == 3 and \
                 self.paragraph['conditions'][0][0] == '# bug437293' and \
                 self.paragraph['conditions'][1].endswith('64bit'):
                self.paragraph['conditions'] = []
            if self._condition_define:
                # If we have define conditions and possible bcond start
                # we need to put it bellow bcond definitions as otherwise
                # the switches do not have any effect
                if self._condition_bcond:
                    self.paragraph['bcond_conditions'] += self.paragraph['conditions']
                elif len(self.paragraph['define']) == 0:
                    self.paragraph['bconds'] += self.paragraph['conditions']
                else:
                    self.paragraph['define'] += self.paragraph['conditions']
                # in case the nested condition contains define we consider all parents
                # to require to be on top too;
                if len(self._oldstore) == 0:
                    self._condition_define = False
            else:
                self.paragraph['build_conditions'] += self.paragraph['conditions']

            # bcond must be reseted when on top and can be set even outside of the
            # define scope. So reset it here always
            if len(self._oldstore) == 0:
                self._condition_bcond = False
            self.paragraph['conditions'] = []

    def _find_pkgconfig_statements(self, listname):
        for i in self.paragraph[listname]:
            if isinstance(i, str):
                if 'pkgconfig(' in i and not self._find_pkgconfig_declarations(listname):
                    return True
            elif isinstance(i, list):
                for j in i:
                    if 'pkgconfig(' in j and not self._find_pkgconfig_declarations(listname):
                        return True
        return False

    def _find_pkgconfig_declarations(self, listname):
        for i in self.paragraph[listname]:
            if isinstance(i, str):
                if 'pkgconfig ' in i or i.endswith('pkgconfig'):
                    return True
            elif isinstance(i, list):
                for j in i:
                    if 'pkgconfig ' in j or j.endswith('pkgconfig'):
                        return True
        return False

    def _end_paragraph(self, needs_license=False):
        lines = []

        # add license to the package if missing and needed
        if needs_license:
            if not self.paragraph['license']:
                self.license = self._fix_license(self.license)
                self._add_line_value_to('license', self.license)

        # Check if we need the pkgconfig
        if not self.br_pkgconfig_required and \
           self._find_pkgconfig_statements('buildrequires'):
            self.br_pkgconfig_required = True
        # only in case we are in main scope
        if not self._oldstore:
            if self.br_pkgconfig_required and not self._find_pkgconfig_declarations('buildrequires'):
                self._add_line_value_to('buildrequires', 'pkgconfig')

        # sort based on category order
        for i in self.categories_order:
            # sort-out within the ordered groups based on the key
            if i in self.categories_with_sorted_package_tokens:
                self.paragraph[i].sort(key=self._sort_helper_key)
                self.paragraph[i] = sort_uniq(self.paragraph[i])
            # sort-out within the ordered groups based on the keyword
            if i in self.categories_with_sorted_keyword_tokens:
                self.paragraph[i].sort(key=self._sort_helper_key)
            for group in self.paragraph[i]:
                lines += self._add_group(group)
        if self.current_group:
            # the current group was not added to any category. It's just some
            # random stuff that should be at the end anyway.
            lines += self._add_group(self.current_group)
            self.current_group = []
        return lines

    def _fix_license(self, value):
        # split using 'or', 'and' and parenthesis, ignore empty strings
        licenses = []
        for a in re.split(r'(\(|\)| and | or (?!later))', value):
            if a != '':
                licenses.append(a)
        if not licenses:
            licenses.append(value)

        for (index, my_license) in enumerate(licenses):
            my_license = self.strip_useless_spaces(my_license)
            my_license = my_license.replace('ORlater', 'or later')
            my_license = my_license.replace('ORsim', 'or similar')
            my_license = my_license.rstrip(';')
            my_license = self.reg.re_license_semicolon.sub(' and ', my_license)
            if my_license in self.license_conversions:
                my_license = self.license_conversions[my_license]
            licenses[index] = my_license

        # create back new string with replaced licenses
        s = ' '.join(licenses).replace("( ", "(").replace(" )", ")")
        return s

    def _split_name_and_version(self, value):
        # split the name and version from the requires element
        if self.reg.re_version_separator.match(value):
            match = self.reg.re_version_separator.match(value)
            pkgname = match.group(1)
            version = match.group(2)
        if not version:
            version = ''
        return pkgname, version

    def _fix_pkgconfig_name(self, value):
        # we just rename pkgconfig names to one unified one working everywhere
        pkgname, version = self._split_name_and_version(value)
        if pkgname == 'pkgconfig(pkg-config)' or \
           pkgname == 'pkg-config':
            # If we have pkgconfig dep in pkgconfig it is nuts, replace it
            return 'pkgconfig{0}'.format(version)
        else:
            return value

    def _pkgname_to_pkgconfig(self, value):
        # we just want the pkgname if we have version string there
        # and for the pkgconfig deps we need to put the version into
        # the braces
        pkgname, version = self._split_name_and_version(value)
        pkgconfig = []
        if pkgname == 'pkgconfig':
            return [value]
        if pkgname not in self.pkgconfig_conversions:
            # first check if the package is in the replacements
            return [value]
        else:
            # first split the pkgconfig data
            pkgconf_list = self.pkgconfig_conversions[pkgname].split()
            # then add each pkgconfig to the list
            # print pkgconf_list
            for j in pkgconf_list:
                pkgconfig.append('pkgconfig({0}){1}'.format(j, version))
        return pkgconfig

    def _pkgname_to_brackety(self, value, name, conversions):
        # we just want the pkgname if we have version string there
        # and for the pkgconfig deps we need to put the version into
        # the braces
        pkgname, version = self._split_name_and_version(value)
        converted = []
        if pkgname not in conversions:
            # first check if the package is in the replacements
            return [value]
        else:
            # first split the data
            convers_list = conversions[pkgname].split()
            # then add each pkgconfig to the list
            # print pkgconf_list
            for j in convers_list:
                converted.append('{0}({1}){2}'.format(name, j, version))
        return converted

    def _fix_list_of_packages(self, value, category):
        # we do fix the package list only if there is no rpm call there on line
        # otherwise print there warning about nicer content and skip
        if self.reg.re_rpm_command.search(value):
            if not self.previous_line.startswith('#') and not self.minimal:
                self.current_group.append('# FIXME: Use %requires_eq macro instead')
            return [value]
        tokens = DependencyParser(value).flat_out()
        # loop over all and do formatting as we can get more deps for one
        expanded = []
        for token in tokens:
            # there is allowed syntax => and =< ; hidious
            token = token.replace('=<', '<=')
            token = token.replace('=>', '>=')
            # we also skip all various rpm-macroed content as it
            # is usually not easy to determine how that should be
            # split
            if token.startswith('%'):
                expanded.append(token)
                continue
            # cleanup whitespace
            token = token.replace(' ', '')
            token = re.sub(r'([<>]=?|=)', r' \1 ', token)
            if not token:
                continue
            # replace pkgconfig name first
            token = self._fix_pkgconfig_name(token)
            # in scriptlets we most probably do not want the converted deps
            if category != 'prereq':
                # here we go with descending priority to find match and replace
                # the strings by some optimistic value of brackety dep
                # priority is based on the first come first serve
                if self.pkgconfig:
                    token = self._pkgname_to_pkgconfig(token)
                # checking if it is not list is simple avoidance of running
                # over already converted values
                if type(token) is not list and self.perl:
                    token = self._pkgname_to_brackety(token, 'perl', self.perl_conversions)
                if type(token) is not list and self.tex:
                    token = self._pkgname_to_brackety(token, 'tex', self.tex_conversions)
                if type(token) is not list and self.cmake:
                    token = self._pkgname_to_brackety(token, 'cmake', self.cmake_conversions)
            if isinstance(token, str):
                expanded.append(token)
            else:
                expanded += token
        # and then sort them :)
        expanded.sort()
        return expanded

    def _add_line_value_to(self, category, value, key=None):
        """
            Change a key-value line, to make sure we have the right spacing.

            Note: since we don't have a key <-> category matching, we need to
            redo one. (Eg: Provides and Obsoletes are in the same category)
        """
        keylen = len('BuildRequires:  ')

        if key:
            pass
        elif category in self.category_to_key:
            key = self.category_to_key[category]
        else:
            raise RpmException('Unhandled category in preamble: %s' % category)

        # append : only if the thing is not known macro
        if not key.startswith('%'):
            key += ':'
        # if the key is already longer then just add one space
        if len(key) >= keylen:
            key += ' '
        # fillup rest of the alignment if key is shorter than muster
        while len(key) < keylen:
            key += ' '

        if category in self.categories_with_package_tokens:
            values = self._fix_list_of_packages(value, category)
        else:
            values = [value]

        for value in values:
            line = key + value
            self._add_line_to(category, line)

    def _add_line_to(self, category, line):
        if self.current_group:
            self.current_group.append(line)
            self.paragraph[category].append(self.current_group)
            self.current_group = []
        else:
            self.paragraph[category].append(line)

        self.previous_line = line

    def add(self, line):
        line = self._complete_cleanup(line)

        # if the line is empty, just skip it, unless keep_space is true
        if not self.keep_space and len(line) == 0:
            return

        # if it is multiline variable then we need to append to previous content
        # also multiline is allowed only for define lines so just cheat and
        # know ahead
        elif self.multiline:
            self._add_line_to('define', line)
            # if it is no longer trailed with backslash stop
            if not line.endswith('\\'):
                self.multiline = False
            return

        # If we match the if else or endif we create subgroup
        # this is basically our class again until we match
        # else where we mark end of paragraph or endif
        # which mark the end of our subclass and that we can
        # return the data to our main class for at-bottom placement
        elif self.reg.re_if.match(line) or self.reg.re_codeblock.match(line):
            self._add_line_to('conditions', line)
            self.condition = True
            # check for possibility of the bcond conditional
            if "%{with" in line or "%{without" in line:
                self._condition_bcond = True
            self.start_subparagraph()
            self.previous_line = line
            return

        elif self.reg.re_else.match(line):
            if self.condition:
                self._add_line_to('conditions', line)
                self.end_subparagraph()
                self.start_subparagraph()
            self.previous_line = line
            return

        elif self.reg.re_endif.match(line) or self.reg.re_endcodeblock.match(line):
            self._add_line_to('conditions', line)
            # Set conditions to false only if we are
            # closing last of the nested ones
            if len(self._oldstore) == 1:
                self.condition = False
            self.end_subparagraph(True)
            self.previous_line = line
            return

        elif self.reg.re_comment.match(line):
            if line or self.previous_line:
                self.current_group.append(line)
                self.previous_line = line
            return

        elif self.reg.re_source.match(line):
            match = self.reg.re_source.match(line)
            self._add_line_value_to('source', match.group(2), key='Source%s' % match.group(1))
            return

        elif self.reg.re_patch.match(line):
            match = self.reg.re_patch.match(line)
            # convert Patch: to Patch0:
            if match.group(2) == '':
                zero = '0'
            else:
                zero = ''
            self._add_line_value_to('patch', match.group(3), key='%sPatch%s%s' % (match.group(1), zero, match.group(2)))
            return

        elif self.reg.re_bcond_with.match(line):
            self._add_line_to('bconds', line)
            return

        elif self.reg.re_define.match(line) or self.reg.re_global.match(line) or self.reg.re_onelinecond.match(line):
            if line.endswith('\\'):
                self.multiline = True
            # if we are kernel and not multiline we need to be at bottom, so
            # lets use misc section, otherwise go for define
            if not self.multiline and line.find("kernel_module") >= 0:
                self._add_line_to('misc', line)
            else:
                self._add_line_to('define', line)
            return

        elif self.reg.re_requires_eq.match(line):
            match = self.reg.re_requires_eq.match(line)
            self._add_line_value_to('requires_eq', match.group(1))
            return

        elif self.reg.re_prereq.match(line):
            match = self.reg.re_prereq.match(line)
            if not self.minimal:
                # add the comment about using proper macro which needs
                # investingaton
                if self.previous_line is not None and not self.previous_line.startswith('#') and not self.previous_line.startswith('PreReq'):
                    self.current_group.append('# FIXME: use proper Requires(pre/post/preun/...)')
            self._add_line_value_to('prereq', match.group(1))
            return

        elif self.reg.re_requires_phase.match(line):
            match = self.reg.re_requires_phase.match(line)
            # Put the requires content properly as key for formatting
            self._add_line_value_to('prereq', match.group(2), key='Requires{0}'.format(match.group(1)))
            return

        elif self.reg.re_provides.match(line):
            match = self.reg.re_provides.match(line)
            self._add_line_value_to('provides_obsoletes', match.group(1), key='Provides')
            return

        elif self.reg.re_obsoletes.match(line):
            match = self.reg.re_obsoletes.match(line)
            self._add_line_value_to('provides_obsoletes', match.group(1), key='Obsoletes')
            return

        elif self.reg.re_buildroot.match(line):
            # we only are fine with buildroot only once
            if len(self.paragraph['buildroot']) == 0:
                self._add_line_value_to('buildroot', '%{_tmppath}/%{name}-%{version}-build')
            return

        elif self.reg.re_license.match(line):
            # first convert the license string to proper format and then append
            match = self.reg.re_license.match(line)
            value = match.groups()[len(match.groups()) - 1]
            value = self._fix_license(value)
            # only store subpkgs if they have different licenses
            if not (type(self).__name__ == 'RpmPackage' and not self.subpkglicense):
                self._add_line_value_to('license', value)
            return

        elif self.reg.re_release.match(line):
            match = self.reg.re_release.match(line)
            value = match.group(1)
            if re.search(r'[a-zA-Z\s]', value):
                self._add_line_value_to('release', value)
            else:
                self._add_line_value_to('release', '0')
            return

        elif self.reg.re_summary_localized.match(line):
            match = self.reg.re_summary_localized.match(line)
            # we need to know what language we need
            language = match.group(1)
            # and what value is there
            content = match.group(2)
            self._add_line_value_to('summary_localized', content, key='Summary{0}'.format(language))
            return

        elif self.reg.re_group.match(line):
            match = self.reg.re_group.match(line)
            value = match.group(1)
            if not self.minimal:
                if self.previous_line and not self.previous_line.startswith('# FIXME') and value not in self.allowed_groups:
                    self.current_group.append('# FIXME: use correct group, see "https://en.opensuse.org/openSUSE:Package_group_guidelines"')
            self._add_line_value_to('group', value)
            return

        # loop for all other matching categories which
        # do not require special attention
        else:
            # cleanup
            for (category, regexp) in self.category_to_clean.items():
                match = regexp.match(line)
                if match:
                    return

            # simple matching
            for (category, regexp) in self.category_to_re.items():
                match = regexp.match(line)
                if match:
                    # instead of matching first group as there is only one,
                    # take the last group
                    # (so I can have more advanced regexp for RPM tags)
                    self._add_line_value_to(category, match.groups()[len(match.groups()) - 1])
                    return

            self._add_line_to('misc', line)

    def output(self, fout, newline=True, new_class=None):
        lines = self._end_paragraph(self.subpkglicense)
        self.lines += lines
        Section.output(self, fout, newline, new_class)
